#include <avr/pgmspace.h>
#include <SoftwareSerial.h>
#include "Arduino.h"
#include "cux.h"


double CUX_timeOfLastTX = 0;
int CUX_HZ = 5;// how many ecu polls per second are allowed
double CUX_timeOfLastMessage = 0;
double CUX_lastRxTime = 0;
double CUX_RxTimeout = 10; 
double startTime = 0;


byte _coolantTemp = 200; //deg in C - 40//  degrees in F = ((_coolantTemp -40)*1.8)+32
unsigned int  _RPM = 25000;  //rpm * 4
byte _MPH = 255;   //128=49.1.   64=24.2
byte _intakeTemp = 90;  //90 showed up as 122F,  changed to 80.  observe delta
byte _TPS = 232;    //was 20,reported 7.84% changed to 10. reported 3.92   


bool CUX_scanRunning = FALSE;

byte CUX_outgoingMessage[3] = {0x14, 0, 0xC3};
bool CUX_outgoingMessageInQueue = FALSE;

byte CUX_outgoingMessageSendPosition = 0;
bool CUX_outgoingMessageCurrentPositionSent = 0;
bool CUX_outgoingMessageCurrentPositionAck = 0;
bool CUX_waitingForResponse = 0;
bool CUX_serialResponseAvailable = 0;
bool CUX_responseDataAvailable = 0;
bool CUX_expectedResponseLength = 0;

//RingBuffer CUX_rxBuffer;
byte CUX_incomingMessage[32];
int CUX_incomingMessageLength = 0;


const int  temperature_adc_to_degrees_f[256]PROGMEM ={
  266, 264, 260, 258, 255, 253, 249, 248, 244, 242, 239, 237, 233, 231, 228, 226,
  222, 221, 219, 217, 215, 213, 212, 210, 208, 206, 204, 203, 201, 199, 197, 195,
  194, 192, 190, 188, 186, 185, 183, 181, 179, 177, 176, 174, 172, 170, 168, 167,
  165, 163, 161, 159, 158, 158, 156, 154, 152, 150, 149, 147, 145, 145, 143, 141,
  140, 138, 138, 136, 136, 134, 132, 132, 131, 131, 129, 129, 127, 125, 125, 123,
  123, 122, 122, 120, 120, 118, 118, 116, 116, 114, 114, 113, 113, 111, 111, 109,
  109, 107, 107, 105, 105, 104, 104, 102, 102, 100, 100,  98,  98,  96,  96,  95,
   95,  95,  93,  93,  91,  91,  89,  89,  87,  87,  86,  86,  84,  84,  82,  82,
   82,  82,  80,  80,  78,  78,  78,  77,  77,  75,  75,  73,  73,  73,  71,  71,
   71,  69,  69,  68,  68,  66,  66,  64,  64,  62,  62,  60,  60,  59,  59,  57,
   57,  57,  55,  55,  53,  53,  51,  51,  50,  50,  48,  48,  46,  46,  44,  44,
   44,  44,  42,  42,  42,  41,  41,  41,  39,  39,  37,  37,  37,  35,  35,  35,
   35,  33,  33,  32,  32,  31,  31,  29,  29,  27,  27,  25,  25,  23,  23,  22,
   22,  22,  20,  20,  18,  18,  18,  16,  16,  14,  14,  13,  13,  13,  11,  11,
   11,   9,   9,   7,   7,   5,   5,   4,   4,   2,   2,   0,   0,  -2,  -2,  -4,
   -4,  -4,  -5,  -5,  -7,  -7,  -7,  -9,  -9,  -9, -11, -11, -11, -13, -13, -13
  };

 bool _pid01_20[32] = {  _pid01, _pid02, _pid03, _pid04, _pid05, _pid06, _pid07, _pid08, _pid09, _pid0A, _pid0B, _pid0C, _pid0D, _pid0E, _pid0F, _pid10,
                        _pid11, _pid12, _pid13, _pid14, _pid15, _pid16, _pid17, _pid18, _pid19, _pid1A, _pid1B, _pid1C, _pid1D, _pid1E, _pid1F, _pid20
                     };


//used for scanloop
bool scanloopSent = 0;
int scanloopPos = 1;


//SoftwareSerial CUX_Serial(10, 11); // RX, TX
#define CUX_Serial Serial

bool CUX_initialize()
{
	
  CUX_Serial.begin(7812);
  CUX_resetWriteBits();
}

//*************************************************************************************
//************************** CUX MESSAGE GENERATON FUNCTIONS **************************
//*************************************************************************************

byte nibbleFlip(byte untweakedNibble)
{
  byte tbyte=0;
  byte tweakedLeftNibble = untweakedNibble % 128 ;
  byte tweakedRightNibble = untweakedNibble - tweakedLeftNibble;
  tweakedLeftNibble>>=4;
  for(int i = 0;i<4;i++)
  {
    tbyte<<=1;
    tbyte+=(tweakedLeftNibble % 2);
    tweakedLeftNibble>>=1;
  }

  tbyte=0;
  tweakedLeftNibble = tbyte;
  for(int i = 0;i<4;i++)
  {
    tbyte<<=1;
    tbyte+=(tweakedRightNibble % 2);
    tweakedRightNibble>>=1;
  }
  tweakedLeftNibble+=tbyte;
  return tweakedLeftNibble;
}

void CUX_constructMessage(byte MSB, byte LSB, byte readLength, bool msgWrite)
{
  byte tbyte = _messageFixedId;
  tbyte <<= 4;
  tbyte += readLength;
  tbyte <<= 2;
  tbyte += MSB >> 6;
  CUX_outgoingMessage[0] = tbyte;
  tbyte = (MSB << 2);
  tbyte += LSB >> 6;
  CUX_outgoingMessage[1] = tbyte;
  if (msgWrite)
    tbyte = _writeMessagePrefix << 6;
  else
    tbyte = _readMessagePrefix << 6;
  tbyte += (LSB % 64);
  CUX_outgoingMessage[2] = tbyte;
  CUX_outgoingMessageInQueue = TRUE;
}

int CUX_setMessageLength(int mLength)
{
  int t_int = mLength;
  if (mLength < 17 && mLength > 0)
    t_int--;
  else
  {
    switch (t_int)
    {
      case 80:
        t_int = 16;
        break;
      case 100:
        t_int = 17;
        break;
      case 400:
        t_int = 18;
        break;
      case 512:
        t_int = 19;
        break;
      default:
        t_int = 0;
        break;
    }
  }
  return t_int;
}

int CUX_getExpectedResponseLength()
{
  int t_int = 1;
  if (CUX_outgoingMessageSendPosition == 2)
  {
    t_int = (CUX_outgoingMessage[0] << 1) >> 3;
    if (t_int < 16)
      t_int++;
    else
    {
      switch (t_int)
      {
        case 16:
          t_int = 80;
          break;
        case 17:
          t_int = 100;
          break;
        case 18:
          t_int = 400;
          break;
        case 19:
          t_int = 512;
          break;
      }
    }
  }
  return t_int;
}

void CUX_resetWriteBits()
{
  CUX_Serial.flush();
  CUX_outgoingMessageSendPosition = 0;
  CUX_outgoingMessageInQueue = FALSE;
  CUX_incomingMessageLength = 0;
  CUX_outgoingMessageCurrentPositionAck = 0;
  CUX_waitingForResponse = FALSE;
  CUX_responseDataAvailable = FALSE;
  CUX_outgoingMessageCurrentPositionSent = FALSE;
  CUX_incomingMessageLength = 0;
}

void CUX_read(double CUX_currentTime) {
  while (CUX_Serial.available())
  {
    byte inByte = CUX_Serial.read();
    CUX_incomingMessage[CUX_incomingMessageLength] = inByte;
    CUX_incomingMessageLength++;
    if (CUX_incomingMessageLength >= CUX_expectedResponseLength)
      CUX_serialResponseAvailable = 1;
    CUX_lastRxTime = CUX_currentTime;
  }
}
void CUX_sendMessage(double CUX_currentTime)
{
  if (!CUX_responseDataAvailable && (CUX_currentTime - CUX_lastRxTime > 1)) 
  {
    if (CUX_outgoingMessageInQueue)
      if (!CUX_waitingForResponse)
      { 
        if (!CUX_outgoingMessageCurrentPositionSent)
        { 
          CUX_Serial.write(CUX_outgoingMessage[CUX_outgoingMessageSendPosition]);
          CUX_outgoingMessageCurrentPositionSent = TRUE;
          CUX_expectedResponseLength = CUX_getExpectedResponseLength();
          CUX_waitingForResponse = TRUE;
        }
      }
      else     //current message sent, waiting for response before moving to next byte
      {
        if (!CUX_outgoingMessageCurrentPositionAck )
        { 
          if (CUX_serialResponseAvailable)
          {
            if (CUX_outgoingMessageSendPosition < 2)
            { 
              if (CUX_incomingMessage[CUX_incomingMessageLength - 1] == CUX_outgoingMessage[CUX_outgoingMessageSendPosition])
              { 
                CUX_outgoingMessageCurrentPositionAck = 1;
                CUX_incomingMessageLength = 0;
                CUX_serialResponseAvailable = 0;
              }
              else
              { 
                CUX_incomingMessageLength = 0;
                CUX_serialResponseAvailable = 0;
              }
            }
            else if (CUX_outgoingMessageSendPosition == 2)
            { 
              CUX_outgoingMessageCurrentPositionAck = 1;
              CUX_responseDataAvailable = 1;
            }
          }
          else if (CUX_currentTime - CUX_lastRxTime > CUX_RxTimeout)  //TIMEOUT
          { 
            CUX_resetWriteBits();
     
          }
        }
        else     //acknowledgement recieved (echo sent byte)  reset ack, clear recieve buffer, and move to next position
        {

          CUX_outgoingMessageCurrentPositionAck = 0;
          CUX_waitingForResponse = FALSE;
          CUX_outgoingMessageSendPosition++; //move to next byte
          CUX_incomingMessageLength = 0;  //reset Recieve buffer position
          CUX_outgoingMessageCurrentPositionSent = FALSE;
          if (CUX_outgoingMessageSendPosition > 2)
          { 
            CUX_outgoingMessageSendPosition = 0;
            CUX_outgoingMessageInQueue = FALSE;
            startTime = CUX_currentTime - startTime;
               
          }
          
        }

      }

  }

}

//*************************************************************************************
//**************************** OBD2 DATA REQUEST FUNCTIONS ****************************
//*************************************************************************************

// GENERATE THE PID AVAILBE LIST (MODE 01 PID 00)
// PLACE 4 BYTES COMPOSED OF 32 PID AVAILABLE FLAGS IN dest
// RETURN LENGTH 4 

int PID0100 (byte *dest)  
{ 
  byte _tbyte;
  for (int i = 0; i < 4; i++)
  {
    _tbyte = 0;
    for (int j = 0; j < 8; j++)
    {
      _tbyte = _tbyte * 2 + _pid01_20[i * 8 + j];
    }
    dest[i] = _tbyte;
    
  }
  return (4);
}

// COOLANT TEMPERATURE (MODE 01 PID 05)
// PLACE 1 BYTE IN dest
// RETURN LENGTH 1

int PID0105(byte * dest)  
{
  dest[0] = _coolantTemp;
  return 1;
}

// RPM (MODE 01 PID 12)
// PLACE 2 BYTES IN dest
// RETURN LENGTH 2

int PID010C(byte *dest) //RPM
{
  dest[0] = _RPM / 256;
  dest[1] = _RPM % 256;
  return 2;
}


// KPH  VEHICLE SPEED (MODE 01 PID 13)
// PLACE 1 BYTE IN dest
// RETURN LENGTH 1

int PID010D(byte * dest) //Kilometers per hour
{
  float kph = _MPH * 0.621371;
  dest[0] = (byte)kph;
  return 1;
}

int PID010F(byte * dest) //intake temp
{
  dest[0] = _intakeTemp;
  return 1;
}

int PID0111(byte * dest) //TPS
{
  dest[0] = _TPS;
  return 1;
}


int PID (byte *_ISO_message, byte _ISO_messageLength, byte *dest) {
  byte _tbyte;
  byte obdCommand[15];
  int obdCommandLength = 0;
  int outputLength = 0;

  for (int i = 0; i < _ISO_messageLength; i++)
  {
    if (i == 0)
      dest[outputLength] = _ISO_message[i] + 0x40;
    else
      dest[outputLength] = _ISO_message[i];
    outputLength++;
  }

  switch (_ISO_message[0])
  {
    case 1:
      {
        if (_ISO_messageLength > 1)
          switch (_ISO_message[1])
          {
            case 0:
              obdCommandLength = PID0100(obdCommand);
              break;
            case 5:
              obdCommandLength = PID0105(obdCommand);
              break;
            case 0xC:
              obdCommandLength = PID010C(obdCommand);
              break;
            case 0xD:
              obdCommandLength = PID010D(obdCommand);
              break;
            case 0xF:
              obdCommandLength = PID010F(obdCommand);
              break;
            case 0x11:
              obdCommandLength = PID0111(obdCommand);
              break;
          }
      }
      break;
  }
  if (obdCommandLength > 0)
  {
    for (int i = 0; i < obdCommandLength; i++) {
      dest[outputLength] = obdCommand[i];
      outputLength++;
    }
    return outputLength;
  }
  else
    return (0);
}
//*************************************************************************************
//***************************** CUX COMM ENGINE FUNCTIONS *****************************
//*************************************************************************************






void CUX_startScanLoop() 
{
  CUX_scanRunning = TRUE;
}

void CUX_stopScanLoop()
{
  CUX_scanRunning = TRUE;
}
bool CUX_isScanRunning()
{
  return CUX_scanRunning;
}


void rpmUpdate()
{
  if(!CUX_outgoingMessageInQueue)
  { 
    byte t_MSB = (_rpmAddress >> 8);
    byte t_LSB = (_rpmAddress << 8) >> 8;
    CUX_constructMessage(t_MSB, t_LSB, CUX_setMessageLength(_rpmLength), 0);
  }
  else
  if (CUX_responseDataAvailable)
  {
    double t_RPM = CUX_incomingMessage[0] + (CUX_incomingMessage[1]* 16);
    if(t_RPM==0xFFFF)
	    t_RPM=0;
    else
	    t_RPM=31250/t_RPM*240;
    _RPM=t_RPM;
    CUX_responseDataAvailable = 0;
    scanloopPos++;
  }
}

void mphUpdate()
{
  if(!CUX_outgoingMessageInQueue)
  {
	  byte t_MSB = (_mphAddress >> 8);
    byte t_LSB = (_mphAddress << 8) >> 8;
    CUX_constructMessage(t_MSB, t_LSB, CUX_setMessageLength(_mphLength), 0);
  }
  else
    if (CUX_responseDataAvailable)
    { 
      _MPH=CUX_incomingMessage[0];
      CUX_responseDataAvailable = 0;
      scanloopPos++;
    }	   
}

void tpsUpdate()
{  
  if(!CUX_outgoingMessageInQueue)
  {
    byte t_MSB = (_tpsAddress >> 8);
    byte t_LSB = (_tpsAddress << 8) >> 8;
    CUX_constructMessage(t_MSB, t_LSB, CUX_setMessageLength(_tpsLength), 0);
  }
  else
    if (CUX_responseDataAvailable)
    { 
      int _tempTPS = CUX_incomingMessage[0]  + (CUX_incomingMessage[1]* 16);
      _tempTPS = _tempTPS/2048;
      _TPS=_tempTPS;
      CUX_responseDataAvailable = 0;
      scanloopPos++;
    }
  }

void coolantUpdate()
{
  if(!CUX_outgoingMessageInQueue)
  {
    byte t_MSB = (_coolantTempAddress >> 8);
    byte t_LSB = (_coolantTempAddress << 8) >> 8;
    CUX_constructMessage(t_MSB, t_LSB, CUX_setMessageLength(_coolantTempLength), 0);
  }
  else
    if (CUX_responseDataAvailable)
    {
      int tbyte1 = byte(CUX_incomingMessage[0]%256);
      int ttemp = (pgm_read_byte_near(temperature_adc_to_degrees_f+tbyte1)-32)/1.8+40;
      _coolantTemp=byte(ttemp);
      CUX_responseDataAvailable = 0;
      scanloopPos++;
    }
}

void scanLoopRun(double CUX_currentTime) 
{
	
 // if(CUX_currentTime > CUX_timeOfLastMessage +( 1000/CUX_HZ))
  {
    switch (scanloopPos)
    {
      case 1:
        rpmUpdate();
        break;
      case 2:
        tpsUpdate();
        break;
      case 3:
        mphUpdate();
        break;
      case 4:
        rpmUpdate();
        break;
      case 5:
        tpsUpdate();
        break; 
      case 6:
        coolantUpdate();
        break;
      default:
        scanloopPos = 1;
        scanloopSent = 0;
        break;
    }
    CUX_timeOfLastMessage = CUX_currentTime;
  }
}


void CUX_update(double currentTime) {
  CUX_sendMessage(currentTime);
  CUX_read(currentTime);
   scanLoopRun(currentTime);
}

