#include "RingBuffer.h"
#include <SoftwareSerial.h>
#include <Arduino.h>
#include "cux.h"
#include "iso9141.h" 

RingBuffer rxBuffer;



int baud = 5; //5 bps default;
double bitSampleTime = 1; //us between bit samples
 
#define ISO_bytelength 8 //8 bits per byte
#define ISO_initResponseByte 0x08
byte ISO_payload[68];      //INCOMING MESSAGE WITHOUT THE HEADERS OR CHECKSUM, TO BE PASSED ON TO INTERPRETER (EX:  0100)
byte ISO_payloadLength = 0;   //LENGTH OF INCOMING MESSAGE
 
bool incomingByte[ISO_bytelength];      //STORAGE FOR EACH BIT READ.
//bool new_data = 0;
bool idleRx = 1; //set to 1 when full byte has been recieved, 0 when bit is < (bytelength + start bit)
int ISO_bitPosition = 0;
byte ISO_buffer[64];
int ISO_buffer_pos = 0;
int ISO_rxPin = 2;
int ISO_txPin = 3;
 
bool RX_end = 0;  //end of ISO transmission, prevents from analyzing mid message
#define ISO_fastBaud 10400
#define ISO_slowBaud  5
 
#define incomingMessagePriorityMask 3 //first 3 bits of header byte 0.  110 is all rover ecu responds to
#define headerModeMask 0 // 0 = 3 byte, 1 = 1 byte;  bit 3 of header byte 0
#define IFRMask  1 //in frame response requirement bit. 0 = response requested, 1 = no response requested bit 4 of header byte 0
#define addressModeMask  0 //1 = physical address, 0 = functional address.  (aka y-bit)
#define outgoingMessagePriority 0x02
#define incomingMessageFunctionalAddressMask 0x6A  //functional address for scan tool
#define incomingMessagePhysicalAddressMask 0xF1 //hand held scan tool 9141-2 physical address
#define outgoingMessageTarget 0x6B  //functional address for scan tool
#define ecuPhysicalAddress 0xE8 //9141-2 physical address for ECU
double ISO_currentTime = 0;
double ISO_timeOfLastMessage = 0;
double ISO_lastRxTime = 0;
double ISO_RxTimeout = 8;
bool ISO_alive = 0; //after init @ 10400 bps, has a message been sent in the last 5 seconds? if not, close connection requiring 5 baud init.
 
byte ISO_Response_Header[] = {1, 1, 1}; //reset during initilaization to above #define values 0x48, 0x6B, 0xE8 typical
byte ISO_Request_Header[] = {0, 0, 0}; //reset during initilaization to above #define values 0x68, 0x6A, 0xF1 typical
 
bool slowinit = false;
bool serialRunning = 0;
byte outgoingMessage[25];
int outgoingMessageLength = 0;
 
int ISO_initStatus = 0 ; // 0 = @ 5bps, no message recieved, 1 = @ 5ps, 5bps message recieved, 2 = @ 10400 no message sent, 3= @10400 message sent, 4= @10400 sync message recieved, 5= ready to run;
 
static const String hextable []  = { "0", "1", "2", "3", "4", "5", "6", "7",
                        "8", "9" ,"A", "B", "C", "D", "E", "F" };
SoftwareSerial ISO_Serial(ISO_rxPin, ISO_txPin); // RX, TX
 
 
ISR(TIMER1_OVF_vect)        // interrupt service routine that wraps a user defined function supplied by attachInterrupt
{
  TCNT1 = 53036;            // preload timer
  if (!idleRx)
    readIncomingBit();
}

void ISO_initialize()
{
	
  setISO_Headers();
  
  slowinit = true;
  noInterrupts();           // disable all interrupts
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 53036;            // preload timer 65536-16MHz/256/2Hz
  TCCR1B |= (1 << CS12);    // 256 prescaler
  TIMSK1 |= (1 << TOIE1);   // enable timer overflow interrupt
  
  interrupts();             // enable all interrupts
  pinMode(ISO_rxPin, INPUT_PULLUP);
  pinMode(ISO_rxPin, INPUT);
  serialStart(5, 2, 3);
  enableRxInt();
 }



void updateKeepAlive()    //updates ISO_alive
{
  if ((ISO_currentTime - ISO_lastRxTime) > 5000)
    ISO_alive = 0;
  else ISO_alive = 1;
}
 
bool ISO_isAlive()    //returns true if message sent in last 5 seconds.
{
  return ISO_alive;
}
 
byte outgoingPriorityType()  //assembles the first byte of the response to the scan tool
{
  byte tbyte = 0;
  tbyte += (outgoingMessagePriority);
  tbyte <<= 1;
  tbyte += headerModeMask;
  tbyte <<= 1;
  tbyte += IFRMask;
  tbyte <<= 1;
  tbyte += addressModeMask;
  tbyte <<= 2;
 
  return tbyte;
}
 
byte incomingPriorityTypeMask() //assembles the expected first byte of the scan tool inquiry
{
  byte tbyte = 0;
  tbyte += (incomingMessagePriorityMask);
  tbyte <<= 1;
  tbyte += headerModeMask;
  tbyte <<= 1;
  tbyte += IFRMask;
  tbyte <<= 1;
  tbyte += addressModeMask;
  tbyte <<= 2;
 
  return tbyte;
}
 
bool ISO_serialWrite(byte message)  //!!!!!modify?
{
  //      Serial.print("SENDING:");
  //     Serial.println(message,HEX);
  if (ISO_Serial.isListening())
    ISO_Serial.stopListening();
  ISO_Serial.write(message);
  ISO_Serial.listen();
 
  return (true);
}
 
void setISO_Headers()
{
  ISO_Request_Header[0] = incomingPriorityTypeMask();
  ISO_Request_Header[1] = incomingMessageFunctionalAddressMask;
  ISO_Request_Header[2] = incomingMessagePhysicalAddressMask;
  ISO_Response_Header[0] = outgoingPriorityType();
  ISO_Response_Header[1] = outgoingMessageTarget;
  ISO_Response_Header[2] = ecuPhysicalAddress;
}
 
int ISO_calcChecksum(byte * message, int messageLength)
{
  byte value_total = 0;
  if (messageLength > 0)
    for (int i = 0; i < messageLength; i++)
      value_total += message[i];
 
  return (value_total);
}
 
int constructOutgoingMessage(byte * rawPIDResponse, int _rawPIDResponseLength)
{
  for (int i = 0; i < 3; i++)
  {
    outgoingMessage[i] = ISO_Response_Header[i];
  
    outgoingMessageLength++;
  }
  for (int i = 0; i < _rawPIDResponseLength; i++)
  {
    outgoingMessage[outgoingMessageLength] = rawPIDResponse[i];
    outgoingMessageLength++;
  }
  outgoingMessage[outgoingMessageLength] = ISO_calcChecksum(outgoingMessage, outgoingMessageLength);
  outgoingMessageLength++;
 
  return outgoingMessageLength;
}
 
bool ISO_validateChecksum(byte * message, int messageLength)
{
  bool validate_pass = false;
  byte value_total = 0;
  if (messageLength > 1)
  {
    for (int i = 0; i < messageLength - 1; i++)
      value_total += message[i];
    if (message[messageLength - 1] == value_total)
      validate_pass = true;
  }

  return (validate_pass);
}
  
void addByteToRxBuffer(byte tbyte)  //PLACE COMPLETED BYTE READ IN TO BUFFER, VERIFIED @ 5BPS
{

  rxBuffer.enqueue(tbyte);
 
  idleRx = 1;
  
}
bool ISO_validateTarget(byte * message, int messageLength, byte * messageHeader)
{
  bool validate_pass = false;
  if (messageLength > sizeof(messageHeader) + 1) //header length + 1 (checksum) + data
  {
    for (int i = 0; i < sizeof(messageHeader) - 1; i++)
      if (messageHeader[i] == message[i])
        validate_pass = true;
      else
        validate_pass = false;
  }
 
  return (validate_pass);
}
 
bool ISO_dataAvailable()//!!!!!!!!!!!!
{
  if (ISO_Serial.available())
  {
    byte inByte = ISO_Serial.read();

    addByteToRxBuffer(inByte);
 
  
    ISO_lastRxTime = ISO_currentTime;
  
    return (true);
  }
  else
    return (false);
}
 
bool respondMS1G()//!!!!!!!!!!!!
{
 
  byte rawPIDResponse[20];
  int rawPIDResponseLength = 0;
  int tresponseLength = 0;
  if (ISO_payloadLength == 2)
    if (ISO_payload[0] == 1)
      switch (ISO_payload[1])
      {
        case 0:
         //  Serial.println("=0100\n\r");
          rawPIDResponseLength = PID0100(rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
          for (int i = 0; i < tresponseLength; i++)
          {
            ISO_serialWrite(outgoingMessage[i]);
            //       Serial.print(hextable[outgoingMessage[i] / 16]);
// Serial.print(hextable[outgoingMessage[i] % 16]);
          }
          break;
 
        case 5:
        //   Serial.println("=0105\n\r");
          rawPIDResponseLength = PID0105(rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
          for (int i = 0; i < tresponseLength; i++)
          {
            ISO_serialWrite(outgoingMessage[i]);
           //    Serial.printlnByte(outgoingMessage[i]);
     //      Serial.print(hextable[outgoingMessage[i] / 16]);
// Serial.print(hextable[outgoingMessage[i] % 16]);
      //           Serial.print(",");
          }
          break;
        case 12:
          rawPIDResponseLength = PID010C(rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
          for (int i = 0; i < tresponseLength; i++){
             //       Serial.print(hextable[outgoingMessage[i] / 16]);
// Serial.print(hextable[outgoingMessage[i] % 16]);
          }
          break;
        case 13:
          rawPIDResponseLength = PID010D(rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
          for (int i = 0; i < tresponseLength; i++)
          {
            ISO_serialWrite(outgoingMessage[i]);
                 //    Serial.print(hextable[outgoingMessage[i] / 16]);
// Serial.print(hextable[outgoingMessage[i] % 16]);
          }
          break;
        case 15:
          rawPIDResponseLength = PID010F(rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
          for (int i = 0; i < tresponseLength; i++){
            ISO_serialWrite(outgoingMessage[i]);
          //           Serial.print(hextable[outgoingMessage[i] / 16]);
// Serial.print(hextable[outgoingMessage[i] % 16]);
          }
          break;
        case 17:
          rawPIDResponseLength = PID0111(rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
          for (int i = 0; i < tresponseLength; i++){
            ISO_serialWrite(outgoingMessage[i]);
       //            Serial.print(hextable[outgoingMessage[i] / 16]);
// Serial.print(hextable[outgoingMessage[i] % 16]);
          }
          break;
        default:
          // if nothing else matches, do the default
          // default is optional
          break;
      }
  outgoingMessageLength = 0;
 
  return TRUE;
}
 
 
bool respondMSG()//!!!!!!!!!!!!
{
 ///  Serial.println("Responding\n\r");
  byte rawPIDResponse[20];
  int rawPIDResponseLength = 0;
  int tresponseLength = 0;
  if (ISO_payloadLength == 2)
    if (ISO_payload[0] == 1)
     
         rawPIDResponseLength = PID(ISO_payload,ISO_payloadLength,rawPIDResponse);
          tresponseLength = constructOutgoingMessage(rawPIDResponse, rawPIDResponseLength);
     
          for (int i = 0; i < tresponseLength; i++){
            ISO_serialWrite(outgoingMessage[i]);
          }
           
       
  outgoingMessageLength = 0;
 
  return TRUE;
}
 
void disableRxInt()
{
  //  Serial.println("disablerxint");
  detachInterrupt(0);
}
 
void enableRxTimer()    //!!!START BIT DETECTED, USE TIMER INTERRUPT TO POLL RX PIN.  THIS ENABLES THE TIMER AND SETS THE PRELOAD TO 1.5X BIT TIME
{
  //  Serial.println("enabeerxtimer");
  noInterrupts();           // disable all interrupts.  Arduino supplied function
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 28000;            // preload timer 65536-16MHz/256/2Hz
  TCCR1B |= (1 << CS12);    // 256 prescaler
  TIMSK1 |= (1 << TOIE1);   // enable timer overflow interrupt
  interrupts();             // enable all interrupts.  Arduino supplied function
}
 
void startbitDetect()
{
  if (idleRx)
  {
    //   Serial.println("CHANGE");
    idleRx = 0;
    disableRxInt();
    enableRxTimer();
  }
}
 
 
void enableRxInt()
{

  attachInterrupt(0, startbitDetect, FALLING);  //INTERRUPT 0 = PIN 2
}
 
 
void readIncomingBit()  //SERIAL READ BIT, VERIFIED @ 5 BPS
{
  //bool incomingByte[10];
  //int ISO_bitPosition = 0;
  incomingByte[ISO_bitPosition] = digitalRead(ISO_rxPin) ^ 1;
  // incomingByte[ISO_bitPosition]=((PIND &&4)>>2)^1;

  ISO_bitPosition++;
  if (ISO_bitPosition >= 8)
  {
    byte tbyte = 0;
    for (int i = 0; i < 8; i++)
      tbyte = tbyte * 2 + incomingByte[7 - i];
    ISO_bitPosition = 0;
    addByteToRxBuffer(tbyte);
    enableRxInt();
  }
}
 
 
 
void disableRxTimer()    //!!!START BIT DETECTED, USE TIMER INTERRUPT TO POLL RX PIN.  THIS ENABLES THE TIMER AND SETS THE PRELOAD TO 1.5X BIT TIME
{
  
  TIMSK1 &= (0 << TOIE1);   // disable timer overflow interrupt
 
}
 
 
void clearBuffer()  // CLEARS INCOMING BUFFER
{
  rxBuffer.clearBuffer();
}
 
void setBitSampleTime(int baudrate)
{
  bitSampleTime = 1000000 / baudrate;
}
 
void setBaud(int baudrate)  //SET BAUD RATE
{
  if (baudrate > 0)
  {
    baud = baudrate;
    setBitSampleTime(baudrate);
  }
}
 
void setPins(int _ISO_rxPin, int _ISO_txPin)
{
  ISO_rxPin = _ISO_rxPin;
  ISO_txPin = _ISO_txPin;
}
 
void startSerial()
{
  serialRunning = TRUE;
}
 
 
 
void serialEnd()
{
  disableRxInt();
  serialRunning = 0;
  disableRxTimer();
}
 
void serialStart()
{
  startSerial();
}
 
void serialStart(int baudrate)
{
  setBaud(baudrate);
  serialStart();
}
 
void serialStart(int baudrate, int _ISO_rxPin, int _ISO_txPin)
{
  setPins(_ISO_rxPin, _ISO_txPin);
  serialStart(baudrate);
}
 

 
void resetTo5Baud()
{
  ISO_Serial.flush();
  ISO_Serial.end();
  ISO_initStatus = 0;
  enableRxInt();
  enableRxTimer();
}
 
byte invertByte(byte inByte)
{
  return (inByte ^ 255);
}
 
void ISO_analyze()
{
  int _bufferLength = 0;
  if(ISO_currentTime - ISO_lastRxTime > ISO_RxTimeout)
  {
       RX_end = 1;
  }
  if(RX_end)
  {
    _bufferLength = rxBuffer.bufferLength();
    for(int i = 0;i<_bufferLength;i++)
    {
      ISO_buffer[i]=rxBuffer.dequeue();
      ISO_buffer_pos++;
    }
    if(ISO_validateChecksum(ISO_buffer,ISO_buffer_pos))
    {
      if(ISO_validateTarget(ISO_buffer,ISO_buffer_pos,ISO_Request_Header))
      {
        ISO_payloadLength = ISO_buffer_pos - sizeof(ISO_Request_Header)-1;
        if(ISO_payloadLength > 0)
        {
          for(int i = 0;i<ISO_payloadLength;i++)
            ISO_payload[i]=ISO_buffer[i+sizeof(ISO_Request_Header)];  //ISO message without checksum or headers
          ISO_timeOfLastMessage = ISO_currentTime;
          ISO_alive = 1;
          respondMSG();
        }
      }
    }
    RX_end = 0;
    ISO_payloadLength = 0;
    ISO_buffer_pos=0;
  }
}
 
void ISO_loop()
{  
  if (ISO_initStatus > 2)
    if (!ISO_isAlive())
    {
      resetTo5Baud();
    }
 
    if (rxBuffer.bufferLength()>4)
      if (ISO_initStatus > 4) //incoming iso 10400 data
      {
        ISO_analyze();
      }
 
    switch (ISO_initStatus)
    {
      case 0:
        if (rxBuffer.bufferLength()>0)
        { 
          if (rxBuffer.dequeue()== initAddress)
          {
          
            ISO_lastRxTime = ISO_currentTime;
            ISO_initStatus++;
          }
        }
      break;
 
      case 1:
        serialEnd();
        ISO_Serial.begin(10400);
        slowinit = false;
        ISO_initStatus++;
      break;
      
      case 2:
        ISO_serialWrite(0x55);
        ISO_serialWrite(ISO_initResponseByte);
        ISO_serialWrite(ISO_initResponseByte);
        ISO_initStatus++;
      break;
      
      case 3:
        if (rxBuffer.bufferLength()>0)
          if (rxBuffer.dequeue() == invertByte(ISO_initResponseByte))
            ISO_initStatus++;
      break;
      
      case 4:
        ISO_serialWrite(invertByte(initAddress));
        ISO_initStatus++;
      break;
      default:
      break;
    }

    if (ISO_initStatus > 1)
      ISO_dataAvailable();
}
  
void ISO_update(double currentTime)  //updates local current time variable, but since this is not a class why?  If this becomes a class, make this a pointer addressthat is set during initialization and get rid of this bullshit code
{
  ISO_currentTime = currentTime;
  ISO_loop();
  updateKeepAlive();
}
